{
  :CreateFrame
  :GetBindingKey
  :InCombatLockdown
  :LibStub
  :MainMenuBar
  :MainMenuBar
  :MultiBarLeft
  :MultiBarRight
  :PetActionBarFrame
  :PetHasActionBar
  :PetHasActionBarFrame
  :RegisterStateDriver
  :ReloadUI
  :SetAlpha
  :SlidingActionBarTexture0
  :SlidingActionBarTexture1
  :UIParent
  :idTweaks
  :idTweaksModule
  :idTweaksOptions
  :idTweaksUtils
} = _G
{
  :TL, :TC, :TR,
  :ML, :MC, :MR,
  :BL, :BC, :BR,
} = idTweaks.anchor
{
  :list,
  :map,
} = idTweaksUtils.table
{
  :Option
  :RangeOption
} = idTweaksOptions

LibActionButton = LibStub "LibActionButton-1.0"

class ActionButton
  new: (name, index, id, header) =>
    -- print "ActionButton#new"

    @name = name
    @index = index
    @id = id
    @header = header

    @create_frame!
    @create_hotkey!

  get_hotkey_name: =>
    -- print "ActionButton#get_hotkey_name"

    "#{@name}#{@id}:LeftClick"

  create_frame: =>
    -- print "ActionButton#create_frame"

    @frame = LibActionButton\CreateButton @id, @name, @frame_header
      outOfRangeColoring: "button"
      tooltip: "enabled"
      showGrid: false
      colors:
        range: { 0.8, 0.1, 0.1 }
        mana: { 0.5, 0.5, 1.0 }
      hideElements:
        macro: false
        hotKey: false
        equipped: false
      keyBoundTarget: false
      clickOnDown: false
      flyoutDirection: "UP"

    @frame\SetState 0, "action", @id

    -- TODO: Get the icon from @frame ?
    _G["#{@name}Icon"]\SetTexCoord 0.08, 0.92, 0.08, 0.92

  create_hotkey: =>
    -- print "ActionButton#create_hotkey"

    @hotkey = @frame\CreateFontString nil, "OVERLAY", "GameTooltipText"
    @hotkey\SetFont @hotkey\GetFont!, 13, "OUTLINE"
    @hotkey\SetVertexColor 0.75, 0.75, 0.75
    @hotkey\ClearAllPoints!
    @hotkey\SetPoint TL, @frame, TL, 0, 0
    @hotkey\SetText GetBindingKey @get_hotkey_name!
    @hotkey\Show!

class ActionBar
  new: (name, first_id, last_id) =>
    -- print "ActionBar#new"

    @name = name
    @first_id = first_id
    @last_id = last_id
    @buttons = {}

    @create_header!
    @create_frame!
    @create_background!

  create_frame: =>
    -- print "ActionBar#create_frame"

    @frame = CreateFrame "Frame", @name, UIParent

    index = 0
    for id = @first_id, @last_id
      @add_button index, id
      index = index + 1

  create_header: =>
    -- print "ActionBar#create_header"

    @header = CreateFrame "Frame", "#{@name}Header", UIParent, "SecureHandlerStateTemplate"
    @header\SetAttribute "_onstate-page", [[
      module:SetAttribute("state", newstate)
      control:ChildUpdate("state", newstate)
    ]]
    RegisterStateDriver @header, "page", "0"

  create_background: =>
    -- print "ActionBar#create_background"

    @background = @frame\CreateTexture "#{@name}Background", "BACKGROUND"
    @background\SetAllPoints @frame
    @background\SetColorTexture 0, 0, 0, 0.5

  set_size: =>
    -- print "ActionBar#set_size"

    {
      :padding
      :spacing
    } = @db
    button_amount = #@buttons
    button_width = @buttons[0]\GetWidth!
    button_height = @buttons[0]\GetHeight!

    @frame\SetWidth button_width * button_amount + spacing * (button_amount - 1) + padding * 2
    @frame\SetHeight button_height + padding * 2

  add_button: (index, id) =>
    -- print "ActionBar#add_button:#{index} #{id}"

    @buttons[index] = ActionButton "#{@name}Button#{index}", index, id, @header

  remove_button: (index) =>
    -- print "ActionBar#remove_button:#{index}"

    -- TODO: Potentially harmful. Can"t remove frames?
    @buttons[index] = nil

class ActionBars extends idTweaksModule
  new: (addon) =>
    -- print "ActionBars#new"

    super addon, "ActionBars", "AceConsole-3.0", "AceEvent-3.0"

    @frames = {}

  -- Note: Don"t need this empty method but leaving it here for reference.
  set_options: =>
    -- print "ActionBars#set_options"

    super!

    @options\add_option RangeOption @, "padding", "Padding", "Padding", {
      min: 0
      max: 50
      step: 1
      bigStep: 1
    }, {}, 0

    @options\add_option RangeOption @, "spacing", "Spacing", "Spacing", {
      min: 0
      max: 50
      step: 1
      bigStep: 1
    }, {}, 0

  -- initialize: =>
  --   -- print "ActionBars#initialize"

  enable: =>
    -- print "ActionBars#enable"

    @disable_blizzard_bars!
    @enable_action_bars!
    @enable_pet_bar!

    @position_all_bars!
    @size_all_bars!

    @register_keybindings!

  disable: =>
    -- print "ActionBars#disable"

    @disable_pet_bar!
    @disable_action_bars!
    @enable_blizzard_bars!

  zoom_button_texture: (texture) =>
    texture\SetTexCoord 0.08, 0.92, 0.08, 0.92

  set_bar_background: (bar) =>
    bar.background = bar\CreateTexture "#{bar.bar_name}Background", "BACKGROUND"
    bar.background\SetAllPoints bar
    bar.background\SetColorTexture 0, 0, 0, 0.5

  get_bar_button_amount: (bar) =>
    bar.last_id + 1 - bar.first_id

  get_button_keybind_name: (bar, button) =>
    "idTweaksActionBar#{bar.id}Button#{button.id}:LeftClick"

  get_pet_button_keybind_name: (button) =>
    "idTweaksActionBarPetButton#{button.id}:LeftClick"

  register_keybindings: =>
    print "ActionBars#register_keybindings"

    _G.BINDING_HEADER_IDTBAR1 = "Bar 1"
    _G.BINDING_HEADER_IDTBAR2 = "Bar 2"
    _G.BINDING_HEADER_IDTBAR3 = "Bar 3"
    _G.BINDING_HEADER_IDTBAR4 = "Bar 4"
    _G.BINDING_HEADER_IDTBAR5 = "Bar 5"
    _G.BINDING_HEADER_IDTPETBAR = "Pet Bar"

    x = 1
    for i = 1, 5
      print "bar#{i}"

      bar = @frames["bar#{i}"]
      
      if not bar
        print "NOT FOUND, #{i}"

      for j = 1, 12
        button = bar["button#{j}"]

        _G["BINDING_NAME_CLICK #{@get_button_keybind_name bar, button}"] = "Bar #{i} Button #{j}"

        x = x + 1

    for j = 1, 10
      button = @frames.petbar["button#{j}"]

      _G["BINDING_NAME_CLICK #{@get_pet_button_keybind_name button}"] = "Pet Bar Button #{j}"

  enable_blizzard_bars: =>
    print "ActionBars#enable_blizzard_bars"

    MainMenuBar\SetScript "OnShow", nil
    MainMenuBar\Show!

    MultiBarLeft\SetScript "OnShow", nil
    MultiBarLeft\Show!

    MultiBarRight\SetScript "OnShow", nil
    MultiBarRight\Show!

  disable_blizzard_bars: =>
    print "ActionBars#disable_blizzard_bars"

    MainMenuBar\SetScript "OnShow", MainMenuBar.Hide
    MainMenuBar\Hide!

    MultiBarLeft\SetScript "OnShow", MultiBarLeft.Hide
    MultiBarLeft\Hide!

    MultiBarRight\SetScript "OnShow", MultiBarRight.Hide
    MultiBarRight\Hide!

  enable_action_bars: =>
    print "ActionBars#enable_action_bars"

    @frames.bar1 = @frames.bar1 or @create_bar "idTweaksActionBar1", 1, 12
    @frames.bar2 = @frames.bar2 or @create_bar "idTweaksActionBar2", 13, 24
    @frames.bar3 = @frames.bar3 or @create_bar "idTweaksActionBar3", 25, 36
    @frames.bar4 = @frames.bar4 or @create_bar "idTweaksActionBar4", 37, 48
    @frames.bar5 = @frames.bar5 or @create_bar "idTweaksActionBar5", 49, 60

  disable_action_bars: =>
    print "ActionBars#disable_action_bars"

    -- Move the frame out of the UI since we can"t destroy it.
    @frames.bar1\ClearAllPoints!
    @frames.bar1\SetPoint BR, UIParent, TL, 0, 0

  show_or_hide_pet_bar: =>
    print "ActionBars#show_or_hide_pet_bar"

    return if InCombatLockdown!

    if PetHasActionBar!
      unless @frames.petbar\IsShown!
        @frames.petbar\Show!
    else
      @frames.petbar\Hide!

  enable_pet_bar: =>
    print "ActionBars#enable_pet_bar"

    @frames.petbar = @frames.petbar or @create_pet_bar "idTweaksActionBarPet", 1, 10

    @ace_module\RegisterEvent "PLAYER_CONTROL_LOST", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PLAYER_CONTROL_GAINED", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PLAYER_FARSIGHT_FOCUS_CHANGED", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "UNIT_PET", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "UNIT_FLAGS", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_BAR_UPDATE", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_BAR_UPDATE_COOLDOWN", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_BAR_SHOWGRID", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_BAR_HIDEGRID", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_BAR_UPDATE_USABLE", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PET_UI_UPDATE", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PLAYER_TARGET_CHANGED", -> @show_or_hide_pet_bar
    @ace_module\RegisterEvent "PLAYER_MOUNT_DISPLAY_CHANGED", -> @show_or_hide_pet_bar

    @show_or_hide_pet_bar!

  disable_pet_bar: =>
    print "ActionBars#disable_pet_bar"

    return unless @frames.petbar

    @ace_module\UnregisterEvent "PLAYER_CONTROL_LOST"
    @ace_module\UnregisterEvent "PLAYER_CONTROL_GAINED"
    @ace_module\UnregisterEvent "PLAYER_FARSIGHT_FOCUS_CHANGED"
    @ace_module\UnregisterEvent "UNIT_PET"
    @ace_module\UnregisterEvent "UNIT_FLAGS"
    @ace_module\UnregisterEvent "PET_BAR_UPDATE"
    @ace_module\UnregisterEvent "PET_BAR_UPDATE_COOLDOWN"
    @ace_module\UnregisterEvent "PET_BAR_SHOWGRID"
    @ace_module\UnregisterEvent "PET_BAR_HIDEGRID"
    @ace_module\UnregisterEvent "PET_BAR_UPDATE_USABLE"
    @ace_module\UnregisterEvent "PET_UI_UPDATE"
    @ace_module\UnregisterEvent "PLAYER_TARGET_CHANGED"
    @ace_module\UnregisterEvent "PLAYER_MOUNT_DISPLAY_CHANGED"

    @show_or_hide_pet_bar!

    PetHasActionBarFrame\EnableMouse "true"
    SlidingActionBarTexture0:SetAlpha 1
    SlidingActionBarTexture1:SetAlpha 1

    for i = @frames.petbar.first_id, @frames.petbar.last_id
      button = _G["PetActionButton#{i}"]
      button\ClearAllPoints!
      button\SetParent PetActionBarFrame
      button\SetPoint BR, PetActionBarFrame, BL, (65 + ((i - 1) * 38)), 0

    @frames.petbar\Hide!

  create_button: (bar, i, button_name, frame_header) =>
    -- print "ActionBars#create_button #{i} #{button_name}"

    button = LibActionButton\CreateButton i, button_name, bar.header,
      outOfRangeColoring: "button"
      tooltip: "enabled"
      showGrid: false
      colors:
        range: { 0.8, 0.1, 0.1 }
        mana: { 0.5, 0.5, 1.0 }
      hideElements:
        macro: false
        hotKey: false
        equipped: false
      keyBoundTarget: false
      clickOnDown: false
      flyoutDirection: "UP"

    @zoom_button_texture _G["#{button_name}Icon"]

    @create_button_hotkey bar, button

    button

  create_button_hotkey: (bar, button, is_pet_button = false) =>
    -- print "ActionBars#create_button_hotkey"

    keybind_name = if is_pet_button then @get_pet_button_keybind_name button else @get_button_keybind_name bar, button

    -- print "ActionBars#create_button_hotkey:keybind_name #{keybind_name}"

    hotkey_text = GetBindingKey keybind_name

    -- print "ActionBars#create_button_hotkey:hotkey_text #{hotkey_text}"

    button.HotKey\Hide!

    -- Create our own
    button.Hotkey = button\CreateFontString nil, "OVERLAY", "GameTooltipText"

    button.Hotkey\SetFont button.Hotkey\GetFont!, 13, "OUTLINE"
    button.Hotkey\SetVertexColor 0.75, 0.75, 0.75
    button.Hotkey\ClearAllPoints!
    button.Hotkey\SetPoint TL, button, TL, 0, 0
    button.Hotkey\SetText hotkey_text
    button.Hotkey\Show!

  create_bar: (bar_name, first_id, last_id) =>
    -- print "ActionBars#create_bar #{bar_name} #{first_id} #{last_id}"

    bar = CreateFrame "Frame", bar_name, UIParent
    bar.header = CreateFrame "Frame", "#{bar_name}Header", UIParent, "SecureHandlerStateTemplate"
    bar.bar_name = bar_name
    bar.button_name = "#{bar_name}Button"
    bar.first_id = first_id
    bar.last_id = last_id

    @set_bar_background bar

    j = 1
    for i = bar.first_id, bar.last_id
      bar["button#{j}"] = @create_button bar, i, "#{bar.button_name}#{j}"
      j = j + 1

    @set_bar_actions bar

    RegisterStateDriver bar.header, "page", "0"

    bar.header\SetAttribute "_onstate-page", [[
      module:SetAttribute("state", newstate)
      control:ChildUpdate("state", newstate)
    ]]

    bar

  create_pet_bar: (bar_name, first_id, last_id) =>
    -- print "ActionBars#create_pet_bar #{bar_name} #{first_id} #{last_id}"

    bar = CreateFrame "Frame", bar_name, UIParent
    bar.header = CreateFrame "Frame", "#{bar_name}Header", UIParent, "SecureHandlerStateTemplate"
    bar.bar_name = bar_name
    bar.button_name = "#{bar_name}Button"
    bar.first_id = first_id
    bar.last_id = last_id

    @set_bar_background bar

    j = 1
    for i = bar.first_id, bar.last_id
      button = _G["PetActionButton#{i}"]
      button_icon = _G["PetActionButton#{i}Icon"]
      button_border = _G["PetActionButton#{i}NormalTexture2"]

      button\ClearAllPoints!
      button\SetParent bar
      button\SetWidth _G.idTweaksActionBar1Button1\GetWidth!
      button\SetHeight _G.idTweaksActionBar1Button1\GetHeight!

      @zoom_button_texture button_icon
      @create_button_hotkey bar, button, true
      button_border\Hide!

      if button\HasScript "OnUpdate"
        button\HookScript "OnUpdate", ->
          if @frames.petbar
            button_border\Hide!

      bar["button#{j}"] = button

      j = j + 1

    PetActionBarFrame\EnableMouse "false"
    SlidingActionBarTexture0\SetAlpha 0
    SlidingActionBarTexture1\SetAlpha 0

    bar

  set_bar_actions: (bar) =>
    -- print "ActionBars#set_bar_actions #{bar.bar_name}"

    for i = 1, @get_bar_button_amount bar
      action = bar.first_id - 1 + i
      -- print "ActionBars#set_bar_actions #{bar.bar_name} button #{i} action #{action}"

      bar["button#{i}"]\SetState 0, "action", action

  size_bar: (bar) =>
    -- print "ActionBars#size_bar #{bar.bar_name}"

    { :padding, :spacing } = @db

    for i = 1, @get_bar_button_amount bar
      button = bar["button#{i}"]

      button\ClearAllPoints!

      if i == 1
        button\SetPoint ML, bar, ML, padding, 0
      else
        button\SetPoint ML, bar["button#{i - 1}"], MR, spacing, 0

    button_amount = @get_bar_button_amount bar
    button_width = bar.button1\GetWidth!
    button_height = bar.button1\GetHeight!

    bar\SetWidth button_width * button_amount + spacing * (button_amount - 1) + padding * 2
    bar\SetHeight button_height + padding * 2

  size_all_bars: =>
    -- print "ActionBars#size_all_bars"

    for i = 1, 5
      @size_bar @frames["bar#{i}"]

    @size_bar @frames.petbar

  position_all_bars: =>
    -- print "ActionBars#position_all_bars"

    { :margin } = @db

    @frames.bar1\ClearAllPoints!
    @frames.bar1\SetPoint BC, UIParent, BC, 0, margin

    @frames.bar2\SetPoint BC, @frames.bar1, TC, 0, margin
    @frames.bar3\SetPoint BC, @frames.bar2, TC, 0, margin
    @frames.bar4\SetPoint BC, @frames.bar3, TC, 0, margin
    @frames.bar5\SetPoint BC, @frames.bar4, TC, 0, margin
    @frames.petbar\SetPoint BC, @frames.bar5, TC, 0, margin

idTweaks\add_module ActionBars idTweaks
